/*
 * Filename      : lru.cc
 * Purpose       : This program provides an efficient implementation
 *                 of LRU list using hash table and buckets.
 * Author        : Lei Yang
 * Date Created  : 11/27/2006
 * Last Modified : 11/27/2006
*/
#include <iostream>
#include <fstream>
#include "lru.h"

using namespace std;

//typedef unsigned long pid_t;
typedef unsigned long address_t;
typedef lru<address_t> lru_list_type;
typedef __gnu_cxx::hash_map<long, unsigned long> hist_map_type;
typedef pair<lru_list_type, hist_map_type> pid_map_pair;
typedef __gnu_cxx::hash_map<pid_t, pid_map_pair> pid_map_type;
typedef pid_map_type::iterator pid_map_iter;
typedef hist_map_type::iterator hist_map_iter;

pid_map_type pid_map;

static void 
gen_pid(pid_t current_pid, address_t phy_addr)
{  
    // lei - find pid in pid_map or insert 
    pid_map_iter pmi = pid_map.find(current_pid);
    if (pmi == pid_map.end())
    {
      // pid not present, add this pid to pid_map
      lru_list_type new_lru_list;
      new_lru_list.insert(phy_addr);

      hist_map_type new_hist_map;
      new_hist_map[-1] += 1;

      pid_map_pair new_pair(new_lru_list, new_hist_map);
      pid_map[current_pid] = new_pair;
    }
    else
    {
      // pid present, check if the address has been accessed
      int lru_pos = ((pmi->second).first).find(phy_addr);
      // address not in LRU list
      if (lru_pos == -1)
      {
        // insert this address to the lru list of this pid
        ((pmi->second).first).insert(phy_addr);
        // update histogram of this pid
        ((pmi->second).second)[-1] += 1;
      }
      else
      {
        // already in the lru list, update histogram
        ((pmi->second).second)[lru_pos] += 1;
      }
    }
}

/*****************************************************************************/
//lei - print the pid map when trace stops
static void
print_pid_map()
{
  ofstream output_file;
  output_file.open("pid_hist.out", ios::app); 
  pid_map_iter pmi = pid_map.begin();
 
  if (pid_map.empty())
  {
    cout << "empty" << endl;
    return;
  }

  for (; pmi != pid_map.end(); pmi++)
  {
    hist_map_iter hmi = ((pmi->second).second).begin();
    output_file << "PID " << pmi->first << endl;
    for (; hmi != ((pmi->second).second).end(); hmi ++)
      output_file << hmi->first << " " << hmi->second << endl;
  }
  output_file.close();
}

/*****************************************************************************/
int main()
{
#if 0
	lru<int> lru_list;

	lru_list.insert(1);
  lru_list.insert(2);
	lru_list.insert(3);
	lru_list.insert(4);
	lru_list.insert(5);
	lru_list.insert(6);
	
  cout << "find 1 " << lru_list.find(1) << endl;
  lru_list.print_buckets();
  cout << "find 1 " << lru_list.find(1) << endl;
  lru_list.print_buckets();
  cout << "find 3 " << lru_list.find(3) << endl;
  lru_list.print_buckets();
  cout << "find 2 " << lru_list.find(2) << endl;
  lru_list.print_buckets();
  //cout << "find 5 " << lru_list.find(5) << endl;
  //lru_list.print_buckets();

	lru_list.insert(7);
  lru_list.insert(8);
	lru_list.insert(9);
	lru_list.insert(10);
	lru_list.insert(11);
	lru_list.insert(12);
  
  cout << "find 1 " << lru_list.find(1) << endl;
  lru_list.print_buckets();
  cout << "find 3 " << lru_list.find(3) << endl;
  lru_list.print_buckets();
  cout << "find 12 " << lru_list.find(12) << endl;
  lru_list.print_buckets();
  cout << "find 9 " << lru_list.find(9) << endl;
  lru_list.print_buckets();
#endif
  
  int pid, addr;
  for (pid = 0; pid < 2; pid++)
  {
    for (addr = 0; addr < 12; addr++)
      gen_pid(pid, addr);
  }
  
  for (pid = 0; pid < 2; pid++)
  {
    for (addr = 11; addr >= 0; addr--)
      gen_pid(pid, addr);
  }
 
  print_pid_map();
  return 1;
}
